/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  |           =====>>>>> 2002 <<<<=====             |
  |                                                 |
  \-------------------------------------------------/
  
  ===================================================
  
  Name:       ccm.C
  
  @Copyright: Damian J. J. Farnell
  
  Date:       Wed Sep 25 15:32:06 BST 2002
  
  Purpose:    Determine and solve CCM equations.
  
  ===================================================
  
  ---------------------------------------------------

  MODULE NAME: CCM.C 

  PURPOSE:     MAIN ROUTINE WHICH CALLS EACH MODULE

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

// #include <iostream.h>
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>

#include "../include/nr.h"
#include "../include/nrutil.h"
#include "../include/Constants.h"
#include "../include/GlobalVariables.h"
#include "../include/FunctionPrototypes.h"
#include "version.h"

#if (PARALLEL_FLAG)
#include<mpi.h>
#endif

// Run in parallel with:

// mpirun -n 1 -p ./ccm.x 

main(int argc, char *argv[])
{
  
  // int i;
  int run_option=7;  /* default = bit 0+1+2 = 1+2+4 */
  int  t2;   /* measure speed */

#if (PARALLEL_FLAG)

  MPI_Init(&argc, &argv);
  MPI_Comm_size(MPI_COMM_WORLD,&npe);
  MPI_Comm_rank(MPI_COMM_WORLD,&myid);

  if (!myid)  // reduce output
    printf("NPE = %d MYID = %d\n", npe, myid);

#else

  npe=1;
  myid=0;

#endif

  /* ccm.C: 

     Purpose: To enumerate all CCM spin cluster configurations within 
     an LSUBm (or other) approach and then determine and solve the
     CCM ground ket- and bra-state equations.

     ====> The Code Is Split Into 5 Modules: <====

     1) Input.C

     ..(a) Read In Control Variables and Initialise Data.

     ..(b) Define the crystallographic lattice.

     2) Define.C

     ..(a) Enumerate the connected clusters.

     ..(b) Determine the unconnected clsuters.

     ..(c) Choose a subset of these clusters (e.g., those
     ......which conserve s_T^z).

     3) Obtain.C

     ..(a) Loop over total number of fundamental configs.
     
     ..(b) Determine the CCM equations for each `term.'
  
     ..(c) Save CCM equations to disk.

     4. Collect.C

     ..(a) Collects like terms together and thus reduces disk usage.

     5. Solve.C

     ..(a) Solve ket- and bra-state equations.
     ..(b) Find ground-state energy, sublattice magnetisation, etc.

  */

  seconds();  /* initialize unix timer */

  terms_fp=NULL;

  // this is info for debugging and better user feedback
  if (!myid) printf("CCM for Quantum Spin Systems, version " Version "\n");
  if (!myid) printf("int=%d long=%d size_t=%d off_t=%d\n",
             8*(int)sizeof(int),    8*(int)sizeof(long),
             8*(int)sizeof(size_t), 8*(int)sizeof(off_t));
  if (!myid) {
#ifdef __GNUC__
    printf("Compiler: GNUC %d.%d\n",__GNUC__,__GNUC_MINOR__);
#endif
#ifdef __DECCXX_VER
    printf("Compiler: DECCXX %d\n",__DECCXX_VER);
#endif
#ifdef __SUNPRO_CC
    printf("Compiler: SUNPRO_CC 0x%x\n",__SUNPRO_CC);
#endif
#if defined(__mips) && defined(_COMPILER_VERSION)
    printf("Compiler: MIPS_CC %d\n",_COMPILER_VERSION);
#endif
  }

  if (argc == 3) {
    int i;
    run_option=0;               // we convert run_option to bitpattern
    for (i=0;argv[2][i];i++) {  // "23" -> 6, "123" = "4" -> 7, "1" -> 1
      switch(argv[2][i]) {      // bit0 for part1, bit1 for part2, ...
        case '1': run_option|=1; break;
        case '2': run_option|=2; break;
        case '3': run_option|=4; break;
        case '4': run_option|=7; break;
        default: printf("Warning: invalid run_option ignored\n");
      }
    }
    if (!myid)
      printf("Run option is =%s converted to %d\n", argv[2], run_option);
  } else {
    printf("Arguments: input_script_name.ccm run_option_1234\nExiting!!\n");
    exit(1);
  }

  fflush(stdout);
  /* dont buffer output to get all outputs immediately */
  setvbuf(stdout,NULL,_IONBF,0);
  /* be sure all threads have runoption set */
#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  input(run_option, argv[1]);

#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD);
#endif

  if (run_option & 1) define();

#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD);
#endif
  t2 = seconds();        /* get unix time in seconds */
  if (!myid)
    printf("Time: define finished after %ds\n",t2);

#ifndef USE_FILE
  if ((run_option&6)==2 && myid==0)
     { printf("\nWarning: Terms will not be saved! try USE_FILE!\n"); }
  if ((run_option&6)==4 && myid==0)
     { printf("\nError: No term file available! try USE_FILE!\n"); exit(1); }
#else
  if ((run_option&6)==6 && myid==0)
     { printf("\nWarning: undef USE_FILE may be much faster\n"); }
#endif

  if (run_option & 2) {
    obtain();  /* get and store collected CCM terms to data.terms */
    /* ToDo: obtain is badly balanced 
     *    we should start with slowest term and put the next easier term to
     *    the next free mpi-jobs until all jobs are done
     *    output 3 biggest and smalles number of terms per cluster? 
     */
  }
  /* "OBTAIN ... SUCCESSFULLY COMPLETED" outputted */
  t2=seconds();        /* get unix time in seconds */
#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD);  /* 100% CPU, nice 19 */
#endif
  if (!myid)
    printf("Time: obtain finished after %ds\n",t2);
#if (PARALLEL_FLAG) 
  MPI_Barrier(MPI_COMM_WORLD); /* wait for other outputs */
#endif
  

  if (run_option & 4) 
    solve();   /* read data.terms (often) and output data.output */

  t2=seconds();        /* get unix time in seconds */
#if (PARALLEL_FLAG)
  MPI_Finalize();
#endif
  if (!myid)
    printf("Time: solve finished after %ds\n",t2);

  if (terms_fp) {
    if (fclose(terms_fp)) 
      printf("%d: Could not close term file\n",myid);
    terms_fp=0; 
  }
  if (fp_buf) free(fp_buf);

  // if (myid==0)
  //  printf("\nPress a key to close this window.\n\n");
  //  getchar();

}


