/*
  /-------------------------------------------------\
  |                                                 |
  |                                                 |
  |     /=====\     /=====\     |\          /|      |
  |    /       \   /       \    | \        / |      |
  |    |           |            |  \      /  |      |
  |    |           |            |   \    /   |      |
  |    \       /   \       /    |    \  /    |      |
  |     \=====/     \=====/     |     \/     |      |
  |                                                 |
  |                                                 |
  \-------------------------------------------------/

  ===================================================

  Name:       CCM (define.C)
  
  Copyright:  Damian J. J. Farnell

  Date:       Wed Sep 29 1999
 
  Purpose:    Determine and solve CCM LSUBm equations.

  Updated:    Joerg Schulenburg 2005

  ===================================================

  ---------------------------------------------------

  MODULE NAME: DEFINE.C 

  PURPOSE:     Defines the LSUBm clusters recursively.
               Not optimised. 

  ---------------------------------------------------

  This program is free software; you can redistribute it and/or
  modify it under the terms of the GNU General Public License
  as published by the Free Software Foundation; either version 2
  of the License, or (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/

//#include <iostream.h> /* gcc doesn like this, its obsolete */
#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#if (PARALLEL_FLAG>0)
#include <mpi.h>
#endif

#include"../include/nr.h"
#include"../include/nrutil.h"
#include"../include/Constants.h"
#include"../include/Externals.h"
#include"../include/FunctionPrototypes.h"

#define MAX_out 100


int subsetClusters(int **config, int length);
int numSubClusters(int **config, int length);

void define(void)
{

  /* this module defines all of the cluster configurations using
     a recursive algorithm. 

     In essence this is a souped-up version of the algorithm used
     in the original ``define configurations'' C program. Here, for
     complex crystallogrphic lattices and higher spin.

  */

  int i,j,tempLatticeSymmetries;
  int cell;
  int NumberConnected=0;
  int NumberDisconnected=0;
  char configurations[80];
  int **configuration,  **tempConfig;
  int length;

  tempConfig=matrix(0,MAX_LSUBM+1,0,5);  // (int **) vector of (int *) may exit

  
  strcpy(configurations,Control.name);
  strcat(configurations,".configs");
  
  configuration=matrix(0,MAX_LSUBM+1,0,5); // (int **) vector of (int *) may exit

  if (Control.output > 0 && myid == 0) {
    printf("\n============================================================\n");
    printf( "             DEFINE CLUSTERS: START ALGORITHM.\n");
    printf("============================================================\n\n");
  }

  /* !! NEW ALGORITHM FOR S>= 1/2.... 30-8-2000 */

  /* PART 1: DETERMINE THE LATTICE ANIMALS */
  /* ToDo: store only the biggest, which are not covered by others?
   * because clusters are always parts of the animals (clusters<=animals)
   */


  NumberLatticeAnimals=0;
  NumberConfigurations=0;

  tempLatticeSymmetries=LatticeSymmetries.number;
  //  LatticeSymmetries.number=1; // switch of symmetries, if not faster Oct05
  // can be done parallel?

  if (myid == 0)
    printf("PART 1: DETERMINE THE LATTICE ANIMALS (using %d symmetries)\n\n",
          LatticeSymmetries.number);

  // ToDo: Unitcell members could be introduced later like s
  //       so animals would live only on unitcells and depend from
  //       the lattice and its symmetry?
  for(cell=1;cell<=UnitCell.number;cell++) {
    length=1;
    configuration[1][1]=0;
    configuration[1][2]=0;
    configuration[1][3]=0;
    configuration[1][4]=cell;      // start site = 1..UC
    /* this includes DSUB-filtering, connected */
    recursiveClusteringFunction(configuration, length,
                        Params.m, // ToDo: old Params.m, new=.n ??? Sep05
			NumberLatticeAnimals);
  }
  if (myid == 0)
    printf("\ndefine: number of lattice animals: %4d\n", NumberLatticeAnimals);


  if (Control.output > 0 && myid == 0) {
    /* PRINT OUT INFORMATION ON ANIMALS */
    /* as an overview we count 1D,2D,3D and 0..length animals */
    int count[MAX_LSUBM+1][3], sum[3], dim, xyz;
    for(j=0;j<MAX_LSUBM+1;j++) count[j][0]=count[j][1]=count[j][2]=0;
    sum[0]=sum[1]=sum[2]=0;

    printf("PRINT OUT INFORMATION ON ANIMALS:\n\n");
    for(i=1;i<=NumberLatticeAnimals;i++) {
      length=findClusterFromDescriptor(latticeAnimals[i].descriptor,
				     tempConfig);
      xyz=0; dim=1;
      for (j=1; j<=length; j++) {
        if (tempConfig[j][1]) xyz|=1;
        if (tempConfig[j][2]) xyz|=2;
        if (tempConfig[j][3]) xyz|=4;
      }
      if (xyz==7)                     dim=3;
      if (xyz==3 || xyz==5 || xyz==6) dim=2;
      if (xyz<=1 || xyz==2 || xyz==4) dim=1; // include Dim=0
      count[length][dim-1]++;
      if (i==10  && Control.output<5)  printf("\n... skip output ...\n\n");
      if (i>=10  && i<NumberLatticeAnimals-10  && Control.output<5)  continue;
      if (i==100 && Control.output<10) printf("\n... skip output ...\n\n");
      if (i>=100 && i<NumberLatticeAnimals-100 && Control.output<10) continue;
      printf("Lattice animal number %4d (l=%d, xyz=%d, dim=%d):\n",
             i, length, xyz, dim);
      printOutDescriptor(latticeAnimals[i].descriptor);
      printOutClusterFromDescriptor(latticeAnimals[i].descriptor);
    }
    printf("number_of_animals (connected):\n"
           " length |  1D     2D         3D |     1D+2D+3D\n"
           "--------+-----------------------+-------------\n");
    for (i=1;i<=Params.m;i++) {
      sum[0]+=count[i][0];
      sum[1]+=count[i][1];
      sum[2]+=count[i][2];
      printf("    %2d  | %3d %6d %10d | %12d\n",
           i,count[i][0],count[i][1],count[i][2],
             count[i][0]+count[i][1]+count[i][2]);
    }
    printf("--------+-----------------------+-------------\n"
           "   Sum: | %3d %6d %10d | %12d\n",
             sum[0],sum[1],sum[2],sum[0]+sum[1]+sum[2]);
  }
   

  if (!myid)
   printf("\nDEFINE LATTICE ANIMALS: COMPLETED SUCCESSFULLY.\n\n");

  /* PART 2: DETERMINE THE FUNDAMTENTAL SPIN CONFIGURATIONS */
  
  if (!myid)
   printf("\nPART 2: DEFINE FUNDAMENTAL SPIN CONFIGURATIONS"
          " (covered by animals).\n\n");

  /* definition of configurations: */
  /* configurations are  animals who's parts are weigthed by 0..2*s */

  LatticeSymmetries.number=tempLatticeSymmetries;
  findFundamentalSpinConfigurations();

  if (!myid)
   printf("\nDEFINE FUND. SPIN CONFIGS.: COMPLETED SUCCESSFULLY.\n\n");

  /* PART 3: PRINT OUT FUNDAMENTAL SPIN CONFIGURATIONS */
  
  if (!myid)
    printf("\nPART 3: PRINT OUT FUNDAMENTAL SPIN CONFIGURATIONS.\n\n"); 
  
  if (Control.output > 0 && myid == 0) {
    int nc;
    for(i=1;i<=NumberConfigurations; i++) {
      length=findClusterFromDescriptor(spinConfigurations[i].descriptor,
                                       configuration);
      // number of clusters which are disconnected from the others
      nc = numSubClusters(configuration, length);
      if (nc>1) NumberDisconnected++;
          else  NumberConnected++;
      if (i==10  && Control.output<5)  printf("\n... skip output ...\n\n");
      if (i>=10  && i<NumberConfigurations-10  && Control.output<5)  continue;
      if (i==100 && Control.output<10) printf("\n... skip output ...\n\n");
      if (i>=100 && i<NumberConfigurations-100 && Control.output<10) continue;
      printf("Cluster Number (%d, parts=%d): \n",i,nc);
      printOutDescriptor(spinConfigurations[i].descriptor);
      printOutClusterFromDescriptor(spinConfigurations[i].descriptor);
      printf("\n");
    }
  }
  if (!myid) {
    printf("\n");
    printf("Number of Connected Clusters    = %8d\n", NumberConnected);
    printf("Number of Disconnected Clusters = %8d\n", NumberDisconnected);
    printf("Total Number of Clusters        = %8d\n", NumberConfigurations);
    printf("\n");
  }

  if (!myid) printf("Write ConfigDescriptors(%d*%d*int) to %s (v060611)\n",
      NumberConfigurations, MAX_SPIN_LIST, configurations);
  fp1=fopen(configurations,"w");  
  if (!fp1) {
    printf("define: Cannot Open %s...\nExiting!!\n\n",configurations);
    exit(1);
  }
  // v060611: to have easier filtering using dd bs=.. skip=.. remove Num
  // pre060611: fwrite(&NumberConfigurations,sizeof(int),1,fp1);
  for(i=1;i<=NumberConfigurations;i++) {
    fwrite(spinConfigurations[i].descriptor,
	   ((MAX_SPIN_LIST)*sizeof(int)), 1, fp1);
  }
  fclose(fp1);

  if (!myid)
    printf("DEFINE CLUSTERS: ALGORITHM SUCCESFULLY COMPLETED!\n\n");

}

// Oct05: new function numSubClusters by Joerg
int numSubClusters(int **config, int length) {

  int flag_changed=0; // flag if nn-matrix is initialized
  int cluster[MAX_LSUBM+1]; // ToDo: I mean SUBN
  int i, j, k, i2, index, num_clusters=length, tmp[5];

  if (length>MAX_LSUBM) { printf("Error: length>MAX_LSUBM\n");exit(1); }
   
  for (i=1; i<=length; i++) cluster[i]=i; // start with each site as cluster

  // ToDo: because of recursiv creation of animals better coding would be
  //       pointer to predecissor + added site
  //       tables[configs]={backpointer+(x,y,z,uc)} and n-steps for decoding!
  for (;;) {
    flag_changed=0; // reset flag
    for(i=1;i<=length-1;i++) {
      index = config[i][4];  // cell
      for(k=i+1; k<=length; k++)
        if (cluster[k]!=cluster[i]) // only different clusters
        if (config[i][1] == config[k][1] &&
            config[i][2] == config[k][2] &&
            config[i][3] == config[k][3] &&
            config[i][4] == config[k][4]) {
        num_clusters--; flag_changed=cluster[k];
        for (i2=1;i2<=length;i2++)
          if (cluster[i2]==flag_changed) { cluster[i2]=cluster[i]; }
      }
      for(j=1;j<=LatticeNeighbours.number[index];j++) {
	
	tmp[0]=config[i][1]+LatticeNeighbours.x[index][j];
	tmp[1]=config[i][2]+LatticeNeighbours.y[index][j];
	tmp[2]=config[i][3]+LatticeNeighbours.z[index][j];
	tmp[3]=             LatticeNeighbours.cell[index][j];
	
	// check that we have a given site.
	
	for(k=i+1; k<=length; k++)
          if (cluster[k]!=cluster[i]) // only different clusters
	  if (tmp[0] == config[k][1] &&
	      tmp[1] == config[k][2] &&
	      tmp[2] == config[k][3] &&
	      tmp[3] == config[k][4]) {
            num_clusters--; flag_changed=cluster[k];
            for (i2=1;i2<=length;i2++)
              if (cluster[i2]==flag_changed) { cluster[i2]=cluster[i]; }
        }
      }
    }
    if (!flag_changed) break;
  } // until no changes
  return num_clusters;
}



// in:   configuration=matrix(0,MAX_LSUBM+1,0,5)={0,0,0,1} initially
//       length=1 initially
//       maxorder=Params.m=m(SUBn-m) distance?
// work: configuration=matrix(0,MAX_LSUBM+1,0,5)
// out:  latticeAnimals[].descriptor via updateDescriptors
//       NumberAnimals  (=NumberLatticeAnimals)
  // ToDo: because of recursiv creation of animals better coding would be
  //       pointer to predecissor + added site
  //       tables[configs]={backpointer+(x,y,z,uc)} and n-steps for decoding!
  //   do it parallel!
void recursiveClusteringFunction(int **configuration, 
		 int length, int maxorder, int &NumberAnimals) {

  /* A simple recursive function that determines all connected configs */

  int i,j,k, descriptorIndex;
  int index, tmp[4];
  static int descriptor[MAX_SPIN_LIST];
  static int n_out=0;
  
  // first check whether we have an error using the `length' of cluster. 

  if (length > maxorder) {
    printf("Error in Recursion: length > maxorder!\nExiting...\n\n");
    exit(1); 
  } 

  // first check if the configuration has been seen before
  
  if (Control.output > 2 && myid == 0 && NumberAnimals < 100) {
    printf("Temporary Connected Cluster (length=%d):", length);
    // Temp = Recursive growing
    for(i=1;i<=length;i++)   // ( x,  y,  z, cell)
      printf(" (%d,%d,%d,%d)",
	     configuration[i][1], configuration[i][2],
	     configuration[i][3], configuration[i][4]);
      printf("\n");
  }
  
  // get descriptor from symmetrized configuration[tempLength]
  // without symmetry: findConfigurationToDescriptor(descriptor, configuration, length);
  //  see bookKeeping.C
  findUniqueDescriptorOfCluster(length, configuration, descriptor);

  if (descriptor[1] == 999999) { // Joerg Oct05
    printf("Error recursivClustering: invalid descriptor\n"
           "Params.btLength (=bounding triangle) < Params.m ???\n");
    exit(1);
  }

  descriptorIndex = findDescriptorIndex2(descriptor, NumberAnimals);
  
  if (Control.output > 2 && myid == 0 && NumberAnimals < 100)
    printf("Has it been previously found? = %d\n\n", descriptorIndex);

  // this is new and only works together with the other changes of Oct05
  if (descriptorIndex) return; // much faster if symmetries are used, Oct05
  
  // allow also length<=maxorder (old was: length==maxorder) Oct05 (faster)
  if ((!descriptorIndex) && (length <= maxorder)) {
    // store descriptor to latticeAnimals[].descriptor
    updateDescriptors2(descriptor,NumberAnimals);
    NumberAnimals++;
    if (Control.output > 1 && myid == 0 && n_out < 100) {
      n_out++;
      printf("Lattice Animal Number %d (l=%d):\n", NumberAnimals, length);
      printOutDescriptor(descriptor);
      printOutClusterFromDescriptor(descriptor);
      printf("\n");
    }
  }

  
  // ToDo: if (length == Params.n) return; // maxorder == Params.n?
  if (length == maxorder) {

    // this is a `dummy case' which just makes sure that we don't
    // go over the maximum LSUBm approximation number adjacent
    // lattice sites for the lattice animals.

  } else {

    // length = 0...MAX_LSUBM, ToDo: check outside maxorder <= MAX_LSUBN
    if (length > MAX_LSUBM) {
       printf("Error overflow: length=%d > MAX_LSUBM\n", length);
       exit(1);
    }
    for(i=1;i<=length;i++) {
      index = configuration[i][4];  // cell
      if (Control.output > 2 && myid == 0 && NumberAnimals < 100) {
        printf("Check %d Neighbours of site(%d):\n",
               LatticeNeighbours.number[index], i);
      }
      for(j=1;j<=LatticeNeighbours.number[index];j++) {
	
	// nearest neighbour is at x,y,z, cell put information
	// into configuration and exit.

	tmp[0]=configuration[i][1]+LatticeNeighbours.x[index][j];
	tmp[1]=configuration[i][2]+LatticeNeighbours.y[index][j];
	tmp[2]=configuration[i][3]+LatticeNeighbours.z[index][j];
	tmp[3]=LatticeNeighbours.cell[index][j];
	
        // if (tmp[0] < 0) continue; // removed for the new version
	
	// check that we have NOT repeated a given site.
	
	for(/*numberFlips=1,*/k=1;k<length+1;k++)
	  if (tmp[0] == configuration[k][1] &&
	      tmp[1] == configuration[k][2] &&
	      tmp[2] == configuration[k][3] &&
	      tmp[3] == configuration[k][4]) break;

        if (k<length+1) continue;

#ifdef DSUB
        // this is an extension suggested by Raymond Bishop and Peggy Li
        // only Clusters with a maximum distance of L(m)^2 is allowed
        //  n DSUBm L^2 animals configs  Eg/N    (sample square Jan06)
        //  2   1   1    2       1       -0.64833148  8xMIPS-250MHz 
        //  4 2=3   2-4  4       2       -0.65311289  8x1s
        //  6   4   5    17      12      -0.66258044  8x5s
        //  8   5   8    34      20      -0.66307028  1x110s terms=2431     
        //  10  6   9    48      43      -0.66511173  1x222s 8obtain=15..93s
        //  12  7   10   170     135     -0.66589305  8x12934s=50t6         
        //  14  8   13   703     831     -0.66703559  62s++ =36t7 (E to high?) 
        //  16  9   16   963     1225    -0.6670114795 1+89s++ =65t7  E9>E8 ???    
        //      8xR10k-250MHz-CPUs: 17..90h sum=254h=8*31.7h efficients=0.35
        //  18  10  17   3620    6874                 61+1147s++               
        //  20  11  18   6827    14084                199+3226s++
        //   m=n bad? better n=m^2 ToDo: what about n!=m?
        // plot [:] "a" u ($5**-.5):7 w lp
        // connectedConfigs  <= Animals (1s) but disconnected can be greater
        static double max_dist2=-1, dist2, x1, y1, z1, x2, y2, z2;
        // m is used otherwise, so we redefine it in a useful manner
        const int L2[26]={0,1,1,2,2,4,5,5,8,8,9,9,10,10,13,13,16,16,
                          17,17,18,18,20,20};
               // L(m/2=1..7) replaced later
        if (Params.m > 25) {
          printf("L(m) not defined for m>25. Abort.\n");exit(1); }
        if (max_dist2 < 0 && myid == 0)
          printf(" ==> using DSUBm filter algorithm, L^2=%d <==\n",
                 L2[Params.m]);
        findCartesiansFromGeneralisedPosition(x1,y1,z1,tmp-1);
        for (k=1;k<length+1;k++) {
          findCartesiansFromGeneralisedPosition(x2,y2,z2,configuration[k]);
#ifdef ToDo_Use_BETTER_DSUB_DISTANCE
          dist2 = (x2-x1+1)*(x2-x1+1)
                + (y2-y1+1)*(y2-y1+1)
                + (z2-z1+1)*(z2-z1+1);
#else     /* old definition for compatibility, Peggy, can I remove it? */
          dist2 = (x2-x1)*(x2-x1) + (y2-y1)*(y2-y1) + (z2-z1)*(z2-z1);
#endif
          if (k==1 || (k>1 && max_dist2<dist2)) max_dist2=dist2;
        if (Control.output > 2 && myid == 0 && n_out < 100)
          printf(" k = %2d  dist^2 = %2g  max^2 = %2g"
                 "  %2g %2g %2g  %2g %2g %2g\n",
           k, dist2, max_dist2, x1, y1, z1, x2, y2 ,z2);
        }
        if (max_dist2 > L2[Params.m]+0.1) continue;
        if (Control.output > 2 && myid == 0 && n_out < 100)
          printf(" maxdist^2 = %g, animal belongs to DSUB\n", max_dist2);
#endif

        // add new neighbouring site to the lattice animal
	configuration[length+1][1] = tmp[0];
	configuration[length+1][2] = tmp[1];
	configuration[length+1][3] = tmp[2];
	configuration[length+1][4] = tmp[3];
	
	recursiveClusteringFunction(configuration,length+1,maxorder,
				      NumberAnimals);
      }
    }
  }
}

// Joerg Sep05: simplify code
// compare descriptors
// return: 1 desc1>desc2, 0 desc1==desc2, -1 desc1<desc2
int cmpDescriptors(int *desc1, int *desc2) {
  int i;
  for(i=MAX_SPIN_LIST-1;i>0;i--) {
    if (desc2[i]<desc1[i]) { return  1; } else // greater
    if (desc2[i]>desc1[i]) { return -1; }      // smaller
  }
  return 0; // equal
}

// out: NumberConfigurations, ClusterFile.terms
// called by define() - DEFINE FUNDAMENTAL SPIN CONFIGURATIONS
// old -> see findFundamentalSpinConfigurations(void)
//   start with findRecursiveConfigs(0,config[0][0]);
// CPU propto (Params.m-1)!
// ToDo: disconnected Clusters are only valid, if they ly insight
//       a connected SUBn-m cluster (n-m animal with holes)
//       until know they are counted too (to much for D>1)
//  sample: oo.o is invalid LSUB4, but its valid LSUB5+LSUB6
//          .o..
//          ....
int findRecursiveConfigs(int length, int **config)  // Joerg Nov05
{

  int i, j, uc, nc=0, flips, dist, is_minDescriptor;
  static long int newNumClusters=0, allClusters=0;
  int descriptor[MAX_SPIN_LIST+2];
  int descriptorIndex;
  static int n_out=0;

  if (length>Params.n) return 0;
  if (length==0) {
    printf("\nfindRecursiveConfigs:\n");
    // NumberConfigurations=0;
    for (i=0; i<MAX_BT; i++) {
      config[1][1]=lookUpTable[i][1];
      config[1][2]=lookUpTable[i][2];
      config[1][3]=lookUpTable[i][3];
      if (lookUpTable[i][1]<0) break;
      if (lookUpTable[i][1]>=Params.m) continue;
      if (lookUpTable[i][2]>=Params.m) continue;
      if (lookUpTable[i][3]>=Params.m) continue;
      if (Params.dim==1 && lookUpTable[i][1]>0) continue; /* translation */
      if (Params.dim==2 && lookUpTable[i][2]>0) continue;
      if (Params.dim==3 && lookUpTable[i][3]>0) continue;
      for (uc=1; uc<=UnitCell.number; uc++) { 
        config[1][4]=uc;
        nc += findRecursiveConfigs(1,config);
      }
    }
    // counted without lookup to the config table (Joerg Nov05)
    printf("recursiv NumClusters = %ld : all = %ld\n",
            newNumClusters, allClusters);
    return nc;
  }
  
  for (i=0;i<1;i++) { // just to use break and continue
     // ToDo: better call Subset before calling  slow functions!
     //   call subsetClusters with cluster instead of descriptor!
     if (!subsetClusters(config,length)) break;

     if (Control.output > 2 && myid == 0 && n_out<=100) {
          printf("findRecursiveCluster: len= %d\n", length);
          printOutCluster(length, config);
     }
     // get descriptor from symmetrized config[length]
     //  add 1 if we have already the minimum descriptor
     is_minDescriptor=findUniqueDescriptorOfCluster(length, config,
                                      descriptor);
     newNumClusters+=is_minDescriptor;
     allClusters++;
     
     // Joerg ToDo: if Animals are symmetrized we dont need to symmetrize again
     // without symmetry: findConfigurationToDescriptor(descriptor, config, order);
     descriptorIndex=((is_minDescriptor)?0:-1);
//     descriptorIndex=findDescriptorIndex(descriptor,
//           ((MAX_NUM_CONFIGURATIONS<NumberConfigurations)?
//             MAX_NUM_CONFIGURATIONS:NumberConfigurations));
        
     if (Control.output > 2 && myid == 0 && n_out<=100) {
          printf("newNumClusters= %ld  allClusters= %ld  nc= %d\n",
              newNumClusters, allClusters, nc);  // testwise
          printf("transformed into:\n");
          printOutDescriptor(descriptor);
          printOutClusterFromDescriptor(descriptor);
          printf("Is it a Representor? = %d", is_minDescriptor);
          printf(" : Subset == %d",subsetClusters(config,length));
          printf("\n\n");
     }
        
        
     // also filter out here length > Params.n (flips)
     if ((!descriptorIndex)) {
        
        // if (NumberConfigurations < MAX_NUM_CONFIGURATIONS)
        //  updateDescriptors(descriptor,NumberConfigurations);
        // NumberConfigurations++;
        nc=1;
        if (n_out<100000) n_out++; // output counter
        
        if (Control.output > 1 && myid == 0 && NumberConfigurations<=100) {
          printf("add #config = %d:\n", NumberConfigurations);
          printOutDescriptor(descriptor);
          printOutClusterFromDescriptor(descriptor);
          printf("\n");
          if (NumberConfigurations==100) printf("... skip output ... \n\n");
        }
      }
  }
  for (i=BTarray[config[length][3]]
                [config[length][2]]
                [config[length][1]]; i<MAX_BT; i++) {
      config[length+1][1]=lookUpTable[i][1];
      config[length+1][2]=lookUpTable[i][2];
      config[length+1][3]=lookUpTable[i][3];
      if (lookUpTable[i][1]<0) break;
      if (lookUpTable[i][1]>=Params.m) continue;
      for (uc=1; uc<=UnitCell.number; uc++) { 
        config[length][4]=uc;
        // count num flips for this site, stop if maximum reached
        // maximum distance must be <= Params.m
        // ToDo: sizeM (defined as size of covering connected cluster) must be <= M
        //       can be calculated by numSubClusters too (or integrate here)
        for(flips=0,dist=0,j=1;j<=length;j++) {
          if (config[j][1]==config[length+1][1]
           && config[j][2]==config[length+1][2]
           && config[j][3]==config[length+1][3]
           && config[j][4]==config[length+1][4]) flips++;
          if (abs(config[j][1]-config[length+1][1])>dist) dist=
              abs(config[j][1]-config[length+1][1]);
          if (abs(config[j][2]-config[length+1][2])>dist) dist=
              abs(config[j][2]-config[length+1][2]);
          if (abs(config[j][3]-config[length+1][3])>dist) dist=
              abs(config[j][3]-config[length+1][3]);
        }
        if (flips>=UnitCell.s2[uc]) continue;
        if (dist >=Params.m       ) continue;
        nc += findRecursiveConfigs(length+1,config);
      }
   }
   return nc;

}

// in:  lattice animals (only connected clusters)
// out: NumberConfigurations, ClusterFile.terms
// called by define() - DEFINE FUNDAMENTAL SPIN CONFIGURATIONS
//    - we remove parts of the animals to get disconnected animals
//    - we multiply parts who's spin can be longer then 1/2
// old -> see rekursiveConfigs
// 
void findFundamentalSpinConfigurations(void)
{

  int j, l, animal, is_minDescriptor, nc;
  long int NumberConnected=0, NumberDisconnected=0,
           newNumClusters=0,  allClusters=0;
  // int max_num_spin_flips;
  int descriptor[MAX_SPIN_LIST+2];
  int descriptorIndex;
  int length;
  int **configuration;
  static int spin[MAX_LSUBM+1];  // ToDo: meaning LSUBN+1
  int **latticeAnimal;
  int tempLength;
  int flag, maxNumberFlips;
  static int n_out=0;
  time_t t1, t2;

  time(&t1);        /* get unix time in seconds */
  configuration=imatrix(0,MAX_LSUBM+1,0,5);
//  static int configuration[MAX_LSUBM+1][5];
  latticeAnimal=imatrix(0,MAX_LSUBM+1,0,5);

  // for 1D spin-one this is simple.
  
  // max_num_spin_flips = Params.s2 + 1; // 2*s+1
  
  NumberConfigurations=0;
  
  for(animal=1;animal<=NumberLatticeAnimals;animal++) {
    
    // see bookKeeping.C
    length=findClusterFromDescriptor(latticeAnimals[animal].descriptor,
				     latticeAnimal);

#ifndef DSUB
    if (length<Params.m) continue;
#endif
    // shorter configs build from longest animals, Joerg Nov05

    if (Control.output > 1 && myid == 0 && n_out<=100) {
      printf("findCluster: using Animal[%d] length=%d\n", animal, length);
      if (Control.output > 2) {
        printOutDescriptor(latticeAnimals[animal].descriptor);
        printOutCluster(length, latticeAnimal);
        printf("\n");
      }
    }
    // if (length >  Params.n) continue; // ToDo: speedup using discAnimals

    // Joerg Oct05: we skip small clusters and build shorter from longest
    //     ToDo:    later we dont skip and dont shorten animals
    // if (length <  Params.m) continue;

    if (length > Params.m) { // ToDo: for xxzCubic-8-8 Joerg Sep05, new=.n ?
      printf("Error findFundamentalSpinConfigurations:"
             " length=%d != Params.m=%d for LatticeAnimal=%d"
             "\nExiting!!\n\n", length, Params.m, animal);
      printOutCluster(length, latticeAnimal);
      Control.output=5;
      length=findClusterFromDescriptor(latticeAnimals[animal].descriptor,
                                     latticeAnimal);
      exit(1);
    }

    // slowest part of the program? Joerg
    // loop over (2s+1)^(m)-1 spin states for connected animals
    //  (ToDo: skip all codes having 0 in it, also disconnected animals)
    // set starting spin code to old=00000 new=11111:
    for(int i1=0; i1<length; i1++) spin[i1+1]=0;
    int tmp_n=0; // num flips

    for (;;) {  // loop over (2s+1)^(m)-1 spin states
      for(j=0; j<length; j++) {   // or length for animals length<m
        tmp_n-=spin[j+1];
        spin[j+1]=(spin[j+1]+1)%(Params.s2+1);
        tmp_n+=spin[j+1];
        if (tmp_n>Params.n) { j--; continue; } // skip tmp_n>n
        if (spin[j+1]) break; // continue only on overflow
      } if (j==length) break; // counting finished
      // printf(" %d%d%d%d%d..%d\n",
      //  spin[1],spin[2],spin[3],spin[4],spin[5],spin[Params.m]);

      // spin[1..m]: 001 002 010 011 012 020 021 022 100 ... for s=1
      // ToDo: this is to much! we should split n into (m=length)+1 summands
      //      filtered out in inner functions? (2s)^m -> (m over 2s)?
      // n=2 spin[1..m]: 001 002 010 011 020 100 101 110 200
      //    or           001 010 100 002 011 020 101 110 200

      tempLength=0;
      flag=0;

      for(j=1;j<=length;j++) {
        maxNumberFlips= UnitCell.s2[latticeAnimal[j][4]];
        if (spin[j] > maxNumberFlips) { flag=1; break; } // ignore
        for(l=1;l<=spin[j];l++) { // copy this config 0..2s times
          tempLength++;
          if (tempLength > MAX_LSUBM) continue; // Overflow Error below
          // ToDo: does it make sence? makes program slow
          configuration[tempLength][1] = latticeAnimal[j][1];
          configuration[tempLength][2] = latticeAnimal[j][2];
          configuration[tempLength][3] = latticeAnimal[j][3];
          configuration[tempLength][4] = latticeAnimal[j][4];
        }
      }

      if (flag) continue;

      if (tempLength > MAX_LSUBM) {
        printf("Error findFundamentalSpinConfigurations:"
               " tempLength=%d > MAX_LSUBM\n", tempLength);
        exit(1);
      }
              
      // ToDo: better call Subset before calling  slow functions!
      //   call subsetClusters with cluster instead of descriptor!
      if (!subsetClusters(configuration,tempLength)) continue;

      if (Control.output > 2 && myid == 0 && n_out<=100) {
        printf("findCluster: test spin configuration (Animal[%d] spins=", animal);
        for (int i1=0;i1<length;i1++) printf("%d",spin[i1+1]);
        printf("):\n");
        printOutCluster(tempLength, configuration);
      }
      // ToDo Joerg: is it necessary to check all symmetrized
      //             configs here? Its slow! 
      //         if animals are symmetrized, we can skip symmetry here
      // get descriptor from symmetrized configuration[tempLength]
      //  add 1 if we have already the minimum descriptor
      is_minDescriptor=
        findUniqueDescriptorOfCluster(tempLength, configuration,
                                    descriptor);

      descriptorIndex=findDescriptorIndex(descriptor,
         ((MAX_NUM_CONFIGURATIONS<NumberConfigurations)?
           MAX_NUM_CONFIGURATIONS:NumberConfigurations));

      if (is_minDescriptor && !descriptorIndex) newNumClusters++;
      allClusters++;   // count all checked clusters (measure performance)

      
      if (Control.output > 2 && myid == 0 && n_out<=100) {
        if  (is_minDescriptor && descriptorIndex)
          printf(" Warning: newNumClusters count this config too much\n");
        printf("newNumClusters= %ld  allClusters= %ld  is_minDescriptor= %d"
               " subClusters = %d\n",
            newNumClusters, allClusters, is_minDescriptor, nc);  // testwise
        printf("transformed into:\n");
        printOutDescriptor(descriptor);
        printOutClusterFromDescriptor(descriptor);
        printf("Has it been previously found? = %d", descriptorIndex);
        printf(" : Subset == %d",subsetClusters(configuration,tempLength));
        printf("\n\n");
      }
      
      
      // also filter out here length > Params.n (flips)
      if ((!descriptorIndex)
       // && subsetClusters(configuration,tempLength) // earlier done
          ) {
        // number of clusters which are disconnected from the others
        nc = numSubClusters(configuration, tempLength);
        if (nc>1) NumberDisconnected++;
            else  NumberConnected++;
        if (NumberConfigurations < MAX_NUM_CONFIGURATIONS)
          updateDescriptors(descriptor,NumberConfigurations);
        NumberConfigurations++;
        if (n_out<100000) n_out++; // output counter
        
        if (Control.output > 1 && myid == 0 && NumberConfigurations<=100) {
          printf("add #config = %d:\n", NumberConfigurations);
          printOutDescriptor(descriptor);
          printOutClusterFromDescriptor(descriptor);
          printf("\n");
          if (NumberConfigurations==100) printf("... skip output ... \n\n");
        }
      }
    }
  }
  time(&t2);        /* get unix time in seconds */

  // counted without lookup to the configuration table (Joerg Nov05)
  if(!myid)
    printf("numberConnected = %ld : Disconnected = %ld : Total = %ld\n",
            NumberConnected,  NumberDisconnected, (long)NumberConfigurations);
  if(!myid)
    printf("newNumClusters = %ld : all = %ld : time[s]= %d\n",
            newNumClusters, allClusters, (int)(t2-t1));
  j=-1;
#ifdef EXPERIMENTAL   // defined in Makefile
  // try new method
  j=findRecursiveConfigs(0, configuration);
#endif
  time(&t1);        /* get unix time in seconds */
  if(!myid)
    printf("recursiveClusters = %d : time[s]= %d\n", j, (int)(t1-t2));

  free_imatrix(configuration,0,MAX_LSUBM+1,0,5);
  free_imatrix(latticeAnimal,0,MAX_LSUBM+1,0,5);
  if (NumberConfigurations > MAX_NUM_CONFIGURATIONS) {
    printf("Error: MAX_NUM_CONFIGURATIONS excited, NumConfigs <= %d\n",
            NumberConfigurations);
    exit(1);
  }
}

void obtainDisconnectedClusters(int &NumberDisconnected, 
				int NumberConnected)
{
}

void updateDescriptors(int *descriptor,
		       int numberElements)
{

  int i,j;
  int position;
  int tempLogicFlag;
  
  // first find where to put the data element 
  // just use a simple backwards linear search

  // although this bit is as cumbersome as old routine it is called
  // less often and all of the elements are always sorted!

  if (numberElements >= MAX_NUM_CONFIGURATIONS-1) {
    if (numberElements == MAX_NUM_CONFIGURATIONS-1)
      printf("Error updateDescriptors: MAX_NUM_CONFIGURATIONS excited\n");
    return;
  }

  position=1;

  if (numberElements > 0) {
    for(j=numberElements;j>0;j--) {
      tempLogicFlag=cmpDescriptors(descriptor,
             spinConfigurations[j].descriptor);
      if (tempLogicFlag==1) {
	position=j+1;
	break;
      }
    }
  }
  
  // now put the element in the correct position
  
  if (numberElements > 0) {
    for(j=numberElements;j>=position;j--) {
      for(i=1;i<MAX_SPIN_LIST;i++) {
	spinConfigurations[j+1].descriptor[i]=
	  spinConfigurations[j].descriptor[i];
      }
    }
  }
  
  for(i=1;i<MAX_SPIN_LIST;i++)
    spinConfigurations[position].descriptor[i]=descriptor[i];
  
#if DEBUG
  if (Control.output > 3 && myid == 0) {
    for(j=1;j<=numberElements+1;j++) {
      for(i=1;i<MAX_SPIN_LIST;i++)
	printf("%d ", spinConfigurations[j].descriptor[i]);
      printf("\n");
    }
    printf("\n");
  }
#endif
}

// store descriptor to sorted latticeAnimals[].descriptor
// ToDo: add a hashing table? to make it faster?
// ToDo: search table and insert if not found, return if found
void updateDescriptors2(int *descriptor,
			int numberElements)
{
  
  int i,j;
  int position;
  int tempLogicFlag;

  if (numberElements+1>=MAX_NUM_CONFIGURATIONS) {  // Oct05
    printf("Error updateDescriptors2: NumAnimals>MAX_NUM_CONFIGURATIONS\n");
    exit(1);
  }
  
  // first find where to put the data element 
  // just use a simple backwards linear search

  // although this bit is as cumbersome as old routine it is called
  // less often and all of the elements are always sorted!

  position=1;

  if (numberElements > 0) {
    for(j=numberElements;j>0;j--) {
      tempLogicFlag=cmpDescriptors(descriptor,latticeAnimals[j].descriptor);
      if (tempLogicFlag==1) {
	position=j+1;
	break;
      }
    }
  }
  
  // now put the element in the correct position
  
  // if (numberElements > 0) {
    for(j=numberElements;j>=position;j--) { // backwards is slow (no cache)
      for(i=1;i<MAX_SPIN_LIST;i++) {
	latticeAnimals[j+1].descriptor[i]=
        latticeAnimals[j  ].descriptor[i];
      }
    }
  // }
  
  for(i=1;i<MAX_SPIN_LIST;i++)
    latticeAnimals[position].descriptor[i]=descriptor[i];
  
#if DEBUG
  if (Control.output > 3 && myid == 0) {
    for(j=1;j<=numberElements+1;j++) {
      for(i=1;i<MAX_SPIN_LIST;i++)
	printf("%d ", latticeAnimals[j].descriptor[i]);
      printf("\n");
    }
    printf("\n");
  }
#endif
}

// filter out unwished states
// called by findFundamentalSpinConfigurations()
// Oct05: use cluster + length instead of descriptor as argument!
// ToDo: **int -> (*int)[5]
int subsetClusters(int **config, int length)
{

  int i;
  // int length;
  // int **config;
  int flag;
  double x, y, z,stz;

  // static int **config=imatrix(0,MAX_LSUBM+1,0,5); // slow, do it once!

  // this routine is somewhat lattice specific and is used to 
  // restrict the clusters used to within some well-defined
  // subset.

  // Hence, this subroutine is application-dependent and to be 
  // written by the user.

  // it returns a value of 0 if the cluster is acceptable
  // and a non-zero integer otherwise.

  // length=findClusterFromDescriptor(descriptor,config);

  flag=1;

  // always check that we don't have more than n spin flips.

  if (length > Params.n) flag=0; 

  // restriction 1: NO RESTRICTION!

  // restriction 2: even number of spin flips.

  if (restrictionClusters==2 && flag==1) {
      if (length%2!=0) flag=0;
  }
  
  // restriction 3: S_T^z=0

  if (restrictionClusters==3 && flag==1) {

    // USER DEFINED OPTION!!

    // ADVANCED: USER MUST ALTER CODE IN THIS CASE!!!

    // EXAMPLE: XXZ model or Heisenberg linear-chain, square or cubic
    // lattices arbitrary `s' though.

    // EXAMPLE: Ca V_4 O_9 Neel model state. Take out comment markers
    // on `x+=0.5;' and `y+=0.5;' and recompile

    stz=0.0;
    for(i=1;i<=length;i++) {
      findCartesiansFromGeneralisedPosition(x,y,z,config[i]);
      // for CAVO applications and SRCUBO now too!

         x+=0.5000000;
         y+=0.5000000;

      if (((int)(((fabs(x)+fabs(y)+fabs(z))+0.001)/2.0)) == 
          ((int)(((fabs(x)+fabs(y)+fabs(z))+1.001)/2.0)))
      // For CAVO with striped model state 
      //	if (((int)((fabs(y)+0.001)/2.0))
      //         == ((int)((fabs(y)+1.001)/2.0)))
        stz++;
      else
        stz--;
    }
    if (fabs(stz)>0.1) flag=0;
  }

  // free_imatrix(config,0,MAX_LSUBM+1,0,5);
  return flag;

}


